(***********************************************************************)
(*                              Project Couverture                     *)
(*                                                                     *)
(* file: coverMCDC.ml                                                  *)
(* authors: Adrien Jonquet, Philippe Wang, Alexis Darrasse             *)
(* licence: CeCIL-B                                                    *)
(***********************************************************************)

open Coverage

(* FIXME: Use a Set rather than a list. *)
let rec add_src src_list src = function
| [] ->  src :: src_list
| h :: t ->
    if src = h then src_list
    else add_src src_list src t

let record_event event_array orig src_list event =
  CoverUtils.relocate_event orig event;
  let src = (CoverUtils.get_pos event).Lexing.pos_fname in
  begin
    try begin
      match event_array.((event.Covevent.covev_pos) / 4) with
      | [] -> event_array.((event.Covevent.covev_pos) / 4) <- [event]
      | l -> event_array.((event.Covevent.covev_pos) / 4) <- event ::l
    end with
    | Invalid_argument s -> failwith (s^"    wrong orig")
    | e -> raise e
  end;
  add_src src_list src src_list

let filter_ev_list =
  let aux lu event =
    if List.exists 
         (fun a -> ((fst (CoverUtils.get_loc a) = fst (CoverUtils.get_loc event))
                    && (snd (CoverUtils.get_loc a) = snd (CoverUtils.get_loc event))))
         lu
    then lu else (event :: lu) in
  List.fold_left aux []

let record_events event_array data src_list module_name =
  let ci = open_in (module_name ^ ".cmo.cover") in
  let (orig, ev_list) = Marshal.from_channel ci in
  close_in ci;
  if List.length ev_list > 0 then
    List.fold_left (record_event event_array orig) src_list (filter_ev_list ev_list)
  else
    src_list

(* analyse the MC/DC vectors *)
let is_independant vectors cond =
  let is_mcdc_complementary vec_conds1 vec_conds2 =
    let vec_list = List.combine (Array.to_list vec_conds1) (Array.to_list vec_conds2) in
    let aux (v1, v2) =
      match (v1.TraceMcdc.cond_value, v2.TraceMcdc.cond_value) with
      | (TraceMcdc.Unknown, _) | (_, TraceMcdc.Unknown) -> true
      | _ ->
          if v1.TraceMcdc.cond.Covevent.covev_loc = cond.Covevent.covev_loc then
            v1.TraceMcdc.cond_value <> v2.TraceMcdc.cond_value
          else
            v1.TraceMcdc.cond_value = v2.TraceMcdc.cond_value in
    List.for_all aux vec_list in
  let rec check_independance vector = function
  | [] -> false
  | t :: q when (is_mcdc_complementary t.TraceMcdc.vec_conds vector.TraceMcdc.vec_conds) ->
      if t.TraceMcdc.vec_dec.TraceMcdc.dec_value <> vector.TraceMcdc.vec_dec.TraceMcdc.dec_value then
        true
      else check_independance vector q
  | t :: q -> check_independance vector q in
  let rec glance_vectors = function
  | [] -> false
  | vect :: l  ->
      if check_independance vect vectors then true else glance_vectors l in
  glance_vectors vectors

let is_mcdc mcdc =
  List.for_all (is_independant mcdc.TraceMcdc.vectors)
    mcdc.TraceMcdc.model.TraceMcdc.mod_conds

let cover_conditions conditions mcdc =
  let aux mod_cond cond =
    if not (is_independant mcdc.TraceMcdc.vectors mod_cond) then
      cond.status <- Uncovered in
  List.iter2 aux mcdc.TraceMcdc.model.TraceMcdc.mod_conds conditions

let cover_mcdc src mcdc_trace =
  let res = ref [] in
  let check_mcdc mcdc =
    let decision =
      let l, l' =
        CoverUtils.get_loc mcdc.TraceMcdc.model.TraceMcdc.mod_dec in
        { event = mcdc.TraceMcdc.model.TraceMcdc.mod_dec;
          cnum_start = l.Lexing.pos_cnum;
          cnum_end = l'.Lexing.pos_cnum;
          status = Uncovered } in
      (* Trivial Case: covering a decisions with not enough vectors to be covered *)
      if (List.length mcdc.TraceMcdc.vectors) <= (List.length mcdc.TraceMcdc.model.TraceMcdc.mod_conds) then
        decision.status <- Uncovered
      else begin
        (* Easy case: covering a simple decision with only one condition *)
        (*            checks if the condition took both values (true and false) *)
        if mcdc.TraceMcdc.model.TraceMcdc.mod_conds = [] then
          begin
            let dec_false = ref false in
            let dec_true = ref false in
            let rec check_dec = function
            | [] -> ()
            | vect :: l ->
                begin
                  match vect.TraceMcdc.vec_dec.TraceMcdc.dec_value with
                  | TraceMcdc.Bool true
                      when vect.TraceMcdc.vec_dec.TraceMcdc.nb_dec > 0 ->
                        dec_true := true
                  | TraceMcdc.Bool false
                      when vect.TraceMcdc.vec_dec.TraceMcdc.nb_dec > 0 ->
                        dec_false := true
                  | _ -> ()
                end;
                check_dec l in
            check_dec  mcdc.TraceMcdc.vectors;
            if !dec_false && !dec_true then
              decision.status <- Covered
            else decision.status <- Uncovered;
            dec_true := false;
            dec_false := false;
          end
        else if is_mcdc mcdc then
          decision.status <- Covered
      end;
      let conditions =
        let tmp = ref [] in
        List.iter
          (fun cond ->
            let l, l' = CoverUtils.get_loc cond in
            tmp := {
              event = cond;
              cnum_start = l.Lexing.pos_cnum;
              cnum_end = l'.Lexing.pos_cnum;
              status = Covered
            } :: !tmp)
          mcdc.TraceMcdc.model.TraceMcdc.mod_conds;
        !tmp in
      cover_conditions (List.rev conditions) mcdc;
      res := (decision :: conditions) @ !res in
    List.iter
      (fun (module_name, mcdc_list) ->
        List.iter
          (fun mcdc ->
             if (CoverUtils.get_pos mcdc.TraceMcdc.model.TraceMcdc.mod_dec).Lexing.pos_fname = src then
               check_mcdc mcdc)
          mcdc_list)
      mcdc_trace;
      !res
