(***********************************************************************)
(*                              Project Couverture                     *)
(*                                                                     *)
(* file: main.ml                                                       *)
(* authors: Adrien Jonquet, Philippe Wang, Alexis Darrasse             *)
(* licence: CeCIL-B                                                    *)
(***********************************************************************)

type cover_status = {
  events : Covevent.coverage_event list array;
  exec_trace_list : bool array list;
  mcdc_trace_list : (string * TraceMcdc.mcdc list) list list;
  mutable mcdc_trace : (string * TraceMcdc.mcdc list) list;
  src_list : string list
}

exception No_mcdc_info

let with_mcdc = ref true

let read_trace_file ac s =
  (* check if the trace exists *)
  if Sys.file_exists s then begin
    (* open traces *)
    let ci = open_in s in
    let res = (Marshal.from_channel ci) in
    close_in ci;
    res :: ac
  end else failwith ("exec trace file "^s^" does not exist")

let init_cover data exec_trace_files mcdc_trace_files =
  let code_section_size = Array.length data.Bytecode_loader.code_section in
  (* initialize the array of coverage events *)
  (* read the execution trace files *)
  let exec_trace_list = List.fold_left read_trace_file [] exec_trace_files in
  (* read the mcdc trace files *)
  let mcdc_trace_list = List.fold_left read_trace_file [] mcdc_trace_files in
  (* build the array of events *)
  let events = Array.make code_section_size ([]:Covevent.coverage_event list) in
  let module_names = CoverUtils.get_module_names () in
  let src_list =
    List.fold_left (CoverMCDC.record_events events data) [] module_names in
  if not (Sys.file_exists "coverage_report/") then
    Unix.mkdir "coverage_report/" 0o755;
  if not (Sys.file_exists "coverage_report/coverage.css") then begin
    let css_fh = open_out "coverage_report/coverage.css" in
    HtmlReport.generate_css css_fh;
    close_out css_fh
  end;
  {
    events = events;
    exec_trace_list = exec_trace_list;
    mcdc_trace_list = mcdc_trace_list;
    mcdc_trace = [];
    src_list = src_list
  }

let cover_object exec_trace_list exec data =
  let is_executed pc = List.exists (fun trace -> trace.(pc)) in
  (* check which instruction has been executed *)
  let exec_trace =
    Array.mapi (fun i -> fun _ -> is_executed i exec_trace_list)
      (List.hd exec_trace_list) in
  (* build the html report *)
  HtmlReport.generate_html_asm exec data exec_trace;
  exec_trace

(* fusion the vectors of the mcdc trace files *)
let fusion_mcdc status =
  let vector_fusion vectors_res vectors =
    let module Fusion =
      Set.Make
        (struct
           type t=TraceMcdc.vector
           let compare = Pervasives.compare
         end) in
    let add_vectors = List.fold_left (fun r e -> Fusion.add e r) in
    Fusion.elements (add_vectors (add_vectors Fusion.empty vectors_res) vectors) in
  let fusion mcdc_res mcdc =
    (* same boolean expression *)
    if mcdc.TraceMcdc.model = mcdc_res.TraceMcdc.model then
      mcdc_res.TraceMcdc.vectors <-
        vector_fusion mcdc.TraceMcdc.vectors mcdc_res.TraceMcdc.vectors in
  let fusion_res mcdc_list  mcdc_res =
    List.iter (fusion mcdc_res) mcdc_list in
  let rec add_to_res mod_name mcdc_list = function
  (* module never encountered -> put directly in the final mcdc res *)
  | [] -> status.mcdc_trace <- (mod_name, mcdc_list) :: status.mcdc_trace
  | (mod_name_res, mcdc_list_res) :: l ->
      if mod_name = mod_name_res then
        (* module encountered -> vectors fusion *)
        List.iter (fusion_res mcdc_list) mcdc_list_res
      else add_to_res mod_name mcdc_list l in
  let add_module (module_name, mcdc_list) =
    add_to_res module_name mcdc_list status.mcdc_trace in
  List.iter (List.iter add_module)

let cover_source_file status exec_trace src =
  (* Expression coverage *)
  let expr_statements, expression_res = CoverExpr.cover_expression src exec_trace status.events in
  let mcdc_statements = if !with_mcdc then CoverMCDC.cover_mcdc src status.mcdc_trace else [] in
  if mcdc_statements = [] then print_endline (src ^ ": No MCDC");
  let statements = mcdc_statements @ expr_statements in
  if !with_mcdc then begin
    let rec aux = function
    | (module_name, mcdc :: l2) :: l1 ->
        let module_name =
          String.capitalize (String.sub src 0 ((String.length src) - 3)) in
        if mcdc.TraceMcdc.model.TraceMcdc.mod_dec.Covevent.covev_module =
             module_name then (module_name, mcdc :: l2)
        else aux l1
    | _ -> raise No_mcdc_info in
    (* HTML report generation *)
    begin try
      HtmlReport.generate_mcdc_tabular src (aux status.mcdc_trace) statements;
      HtmlReport.generate_html true src statements
    with No_mcdc_info -> HtmlReport.generate_html false src statements end
  end else HtmlReport.generate_html false src statements;
  expression_res

(* expression coverage *)
(* MC/DC coverage *)
let cover_source status exec_trace exec exec_trace_files mcdc_trace_files =
  (* mcdc_trace fusion *)
  if !with_mcdc then fusion_mcdc status status.mcdc_trace_list;
  let all_cover_res =
    List.map (cover_source_file status exec_trace)
      (List.filter Sys.file_exists status.src_list) in
  HtmlReport.generate_html_index exec all_cover_res exec_trace_files mcdc_trace_files

let exec_trace_files = ref []
let mcdc_trace_files = ref []

let add_exec_trace s = exec_trace_files := s :: !exec_trace_files
let add_mcdc_trace s = mcdc_trace_files := s :: !mcdc_trace_files

let cmd_args =
  [("-mcdc", Arg.String add_mcdc_trace, "mcdc trace files ");
   ("-trace", Arg.String add_exec_trace, "exec trace files")]

let init_exec s =
  if not (Sys.file_exists s) then failwith "executable file does not exist";
  if !exec_trace_files = [] then failwith "no trace files";
  if !mcdc_trace_files = [] then with_mcdc := false;
  (* load the OCaml bytecode exe *)
  let data = Bytecode_loader.load_file s in
  (* the size in byte of the bytecode *)
  let status = init_cover data !exec_trace_files !mcdc_trace_files in
  let exec_trace = cover_object status.exec_trace_list s data in
  cover_source status exec_trace s !exec_trace_files !mcdc_trace_files;
  exit 0

let _ =
  Arg.parse cmd_args init_exec (Sys.argv.(0)^" [options] executable");
  Arg.usage cmd_args (Sys.argv.(0)^" [options] executable")
