(***********************************************************************)
(*                              Project Couverture                     *)
(*                                                                     *)
(* file: pluginTraceMcdc.ml                                            *)
(* authors: Adrien Jonquet, Mathias Bourgoin                           *)
(* licence: CeCIL-B                                                    *)
(***********************************************************************)

open TraceMcdc
(* module Events = Map.Make(struct type t = int let compare = compare end) *)
module Events = struct
  type 'a t = (int * 'a) list
  let empty = []
  let add k x l = (k, x) :: l
  let iter f = List.iter (fun (k, x) -> f k x)
  let find = List.assoc
  let mem = List.mem_assoc
end

type events = Covevent.coverage_event Events.t

(* TODO: use maps rather than lists. *)
let mcdc_trace = ref ([] : ((int * string) * (mcdc list * events * events)) list)

open Instructions

let nb_vectors = ref 0
let nb_merged = ref 0

let vector_completed vector = vector.vec_dec.nb_dec > 0

let rec merge_vector m vector = function
(* FIXME: do we need to add vector at the end rather than the beggining? *)
    [] -> m.vectors <- List.tl m.vectors @ [vector]
  | v :: q ->
      let b = ref true in
      for i = 0 to Array.length v.vec_conds - 1 do
        if vector.vec_conds.(i).cond <> v.vec_conds.(i).cond then assert false;
        match vector.vec_conds.(i).cond_value, v.vec_conds.(i).cond_value with
        | Unknown, Unknown -> ()
        | Bool true, Bool true -> ()
        | Bool false, Bool false -> ()
        | _ -> b := false
      done;
      if (!b) && (vector.vec_dec.dec_value = v.vec_dec.dec_value) then begin
        v.vec_dec.nb_dec <- v.vec_dec.nb_dec + 1;
        m.vectors <- List.tl m.vectors
      end else merge_vector m vector q

let incr_vec vector =
  vector.nb_vec <- vector.nb_vec + 1;
  vector.vec_dec.nb_dec <- vector.vec_dec.nb_dec + 1

let record_dec dec v m =
  let new_vector b = {
    nb_vec = 0;
    vec_dec = {
      dec_value = Bool b;
      nb_dec = 0;
      dec = m.model.mod_dec
    };
    vec_conds = [||]
  } in
  if m.model.mod_conds = [] then begin
    if m.vectors = [] then m.vectors <- [ new_vector true; new_vector false ];
    incr_vec (List.nth m.vectors (if v then 0 else 1))
  end else
    let vector = List.hd m.vectors in
    vector.vec_dec <- { dec_value = Bool v; nb_dec = 1; dec = m.model.mod_dec };
  if List.length m.vectors > 1 then
    merge_vector m (List.hd m.vectors) (List.tl m.vectors)

let record_cond cond v m =
  let same_loc cond cond' = cond'.Covevent.covev_loc = cond.Covevent.covev_loc in
  if List.mem cond m.model.mod_conds then begin
    if (m.vectors == []) || (cond = List.hd m.model.mod_conds) then
      let new_vec_cond i =
        let eq = same_loc cond (List.nth m.model.mod_conds i) in
        { cond_value = if eq then Bool v else Unknown;
          nb_cond = if eq then 1 else 0;
          cond = List.nth m.model.mod_conds i } in
      m.vectors <- {
        nb_vec = 0;
        vec_dec = {
          dec_value = Unknown;
          nb_dec = 0;
          dec = m.model.mod_dec
        };
        vec_conds = Array.init (List.length m.model.mod_conds) new_vec_cond
      } :: m.vectors
    else
      let vector = List.hd m.vectors in
      for i=0 to (Array.length vector.vec_conds)-1 do
        let vec_cond = vector.vec_conds.(i) in
        if same_loc cond vec_cond.cond then begin
          vec_cond.nb_cond <- vec_cond.nb_cond + 1;
          vec_cond.cond_value <- Bool v
        end
      done
  end

let last_instr = ref ATOM0

let find_cov_from_adr pc l =
  try Some (Events.find pc l) with Not_found -> None

let cond_value vm = function
| BOOLNOT -> vm.Vm.accumulator <> Value.Int 0
| BRANCHIF n -> vm.Vm.accumulator <> Value.Int 0
| BRANCHIFNOT n -> vm.Vm.accumulator <> Value.Int 0
| RETURN n -> vm.Vm.accumulator <> Value.Int 0
| ACC0 -> Vm.peek vm 0 <> Value.Int 0
| ACC1 -> Vm.peek vm 1 <> Value.Int 0
| ACC2 -> Vm.peek vm 2 <> Value.Int 0
| ACC3 -> Vm.peek vm 3 <> Value.Int 0
| ACC4 -> Vm.peek vm 4 <> Value.Int 0
| ACC5 -> Vm.peek vm 5 <> Value.Int 0
| ACC6 -> Vm.peek vm 6 <> Value.Int 0
| ACC7 -> Vm.peek vm 7 <> Value.Int 0
| ACC n -> Vm.peek vm n <> Value.Int 0
| CONST0 -> false
| CONST1 -> true
| _ -> vm.Vm.accumulator <> Value.Int 0

(* record the boolean vectors for MC/DC *)
let record_mcdc vm (mcdc_trace, cond_list, dec_list) instr =
  begin
    match find_cov_from_adr (vm.Vm.code_pointer * 4) cond_list with
    | Some cond -> begin
        match (instr, !last_instr) with
          (RETURN _, BRANCHIF _) | (RETURN _, BRANCHIFNOT _) -> ()
        | _ ->
            let v = cond_value vm instr in
            List.iter (record_cond cond v) mcdc_trace
      end
    | None -> ()
  end;
  begin
    match find_cov_from_adr (vm.Vm.code_pointer * 4) dec_list with
    | Some dec ->
        let v = vm.Vm.accumulator <> Value.Int 0 in
        let m = List.find (fun m -> m.model.mod_dec = dec) mcdc_trace in
        record_dec dec v m
    | None -> ()
  end;
  last_instr := instr

let remove_dups =
  let filter lu event =
    let test a = CoverUtils.get_loc a = CoverUtils.get_loc event in
    if List.exists test lu then lu else (event :: lu) in
  List.fold_left filter []

(* get the list of decisions and the list of conditions *)
let first_pass =
  let aux (min_pos, mcdc_l, cond_l, dec_l) cov =
    let new_min_pos = min min_pos cov.Covevent.covev_pos in
    match cov.Covevent.covev_status with
    | Covevent.Coverage_event_decision ->
        let vect = {
          model = {mod_dec = cov; mod_conds = []};
          vectors = []
        } in
        (new_min_pos, (vect::mcdc_l), cond_l, Events.add cov.Covevent.covev_pos cov dec_l)
    | Covevent.Coverage_event_condition _ ->
        (new_min_pos, mcdc_l, Events.add cov.Covevent.covev_pos cov cond_l, dec_l)
    | _ -> (new_min_pos, mcdc_l, cond_l, dec_l) in
    List.fold_left aux (max_int, [], Events.empty, Events.empty)

let rec second_pass mcdc_list =
  let aux cond dec t =
    if t.model.mod_dec.Covevent.covev_loc = dec.Covevent.covev_loc then
      t.model.mod_conds <- cond :: t.model.mod_conds in
  let populate_vector_model cond dec = List.iter (aux cond dec) mcdc_list in
  let populate_mcdc _ cond = match cond.Covevent.covev_status with
    Covevent.Coverage_event_condition dec -> populate_vector_model cond dec
  | _ -> assert false in
  Events.iter populate_mcdc

(* build the boolean vectors for MC/DC *)
let build_mcdc_vectors data module_name =
  (* get the list of modules *)
  let ci = open_in (module_name ^ ".cmo.cover") in
  let (orig, ev_list) = Marshal.from_channel ci in
  close_in ci;
(*  if ev_list = [] then failwith ("no coverage_events for module " ^ module_name); *)
  List.iter (CoverUtils.relocate_event orig) ev_list;
  let min_pos, mcdc_list, cond_list, dec_list = first_pass (remove_dups ev_list) in
  second_pass mcdc_list cond_list;
  (min_pos, mcdc_list, cond_list, dec_list)

let mcdc_trace_enabled = ref false
let mcdc_trace_name = ref ""

let compare_mod ((pos1, _), _) ((pos2, _), _) = compare pos2 pos1

let init data =
(* TODO: move the vector building in compile-time. *)
  let init_module module_name =
    let min_pos, mcdc_list, cond_list, dec_list = build_mcdc_vectors data module_name in
    ((min_pos, module_name), (mcdc_list, cond_list, dec_list)) in
  if !mcdc_trace_enabled then
    let module_names = CoverUtils.get_module_names () in
    mcdc_trace := List.sort compare_mod (List.map init_module module_names)

let step vm instruction =
  let rec step_module = function
    ((min_pos, _), mcdc_trace_list) :: ls ->
      if vm.Vm.code_pointer * 4 >= min_pos then
        record_mcdc vm mcdc_trace_list instruction
      else step_module ls
  | [] -> () in
  (* record the current obj instruction for source coverage *)
  (* FIXME: apply step_module only for the module including the current instruction. *)
  if !mcdc_trace_enabled then step_module !mcdc_trace

let extract_trace =
  let aux (module_name, (trace, _, _)) = (module_name, trace) in
  List.map aux

let finalise () =
  if !mcdc_trace_enabled then
    (* mcdc trace generation *)
    try begin
      let co = open_out !mcdc_trace_name in
      Marshal.to_channel co (extract_trace !mcdc_trace) [];
      close_out co
    end with
    | Sys_error s -> failwith ("pluginTraceMcdc.ml : "^s)

let _ =
  Plugin.register_plugin init step finalise
    ("-mcdc", Arg.String (fun s -> mcdc_trace_enabled := true; mcdc_trace_name := s),
     "file write mcdc trace in file")
